import asyncio
from contextlib import AsyncExitStack, asynccontextmanager, suppress
from types import TracebackType
from typing import AsyncIterator, Dict, Final, Hashable, Optional, Type

__all__ = ("FreqLimit", "__version__")
__version__ = "0.0.13"


class Lock:
    def __init__(self) -> None:
        self._ts: float = -float("inf")
        self._count: int = 0
        self._lock: Final = asyncio.Lock()

    @property
    def ts(self) -> float:
        return self._ts

    @ts.setter
    def ts(self, ts: float) -> None:
        self._ts = ts

    async def __aenter__(self) -> None:
        self._count += 1
        await self._lock.acquire()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        try:
            self._lock.release()
        finally:
            self._count -= 1

    @property
    def count(self) -> int:
        return self._count


class FreqLimit:
    def __init__(self, interval: float, clean_interval: float = 0) -> None:
        if interval <= 0:
            raise RuntimeError("Interval must be greater than 0")
        if clean_interval < 0:
            raise RuntimeError(
                "Clean interval must be greater than or equal to 0"
            )
        self._interval: Final[float] = interval
        self._clean_interval: Final[float] = (
            clean_interval if clean_interval > 0 else interval
        )
        self._locks: Final[Dict[Hashable, Lock]] = {}
        self._clean_event: Final = asyncio.Event()
        self._clean_task: Optional[asyncio.Task[None]] = None
        self._loop: Final = asyncio.get_running_loop()

    @asynccontextmanager
    async def resource(self, key: Hashable = None) -> AsyncIterator[None]:
        if self._clean_task is None:
            self._clean_task = asyncio.create_task(self._clean())
        if key not in self._locks:
            self._locks[key] = Lock()
        async with AsyncExitStack() as stack:
            stack.callback(self._clean_event.set)
            await stack.enter_async_context(self._locks[key])
            delay = self._interval - self._loop.time() + self._locks[key].ts
            if delay > 0:
                await asyncio.sleep(delay)
            self._locks[key].ts = self._loop.time()
            yield

    async def clear(self) -> None:
        if self._clean_task is not None:
            self._clean_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._clean_task
            self._clean_task = None
        self._locks.clear()
        self._clean_event.clear()

    async def _clean(self) -> None:
        while True:
            if len(self._locks) == 0:
                await self._clean_event.wait()
                self._clean_event.clear()
            for key in tuple(self._locks):
                age = self._loop.time() - self._locks[key].ts
                if self._locks[key].count == 0 and age >= self._clean_interval:
                    del self._locks[key]
            await asyncio.sleep(self._clean_interval)



import re
from pathlib import Path

from setuptools import setup  # type: ignore

path = Path(__file__).parent
txt = (path / "aiofreqlimit" / "__init__.py").read_text("utf-8")
version = re.findall(r"^__version__ = \"([^\"]+)\"\r?$", txt, re.M)[0]
readme = (path / "README.rst").read_text("utf-8")

setup(
    name="aiofreqlimit",
    version=version,
    description="Frequency limit for asyncio",
    long_description=readme,
    long_description_content_type="text/x-rst",
    url="https://github.com/gleb-chipiga/aiofreqlimit",
    license="MIT",
    author="Gleb Chipiga",
    # author_email='',
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Internet",
        "Framework :: AsyncIO",
    ],
    packages=["aiofreqlimit"],
    package_data={"aiofreqlimit": ["py.typed"]},
    python_requires=">=3.9,<3.12",
    tests_require=[
        "pytest",
        "pytest-asyncio",
        "pytest-mock",
        "pytest-cov",
        "hypothesis",
    ],
)



import asyncio
from contextlib import suppress
from random import uniform
from typing import List, Tuple, cast
from weakref import ref

import pytest
from hypothesis import given
from hypothesis.strategies import floats
from pytest_mock import MockerFixture

import aiofreqlimit


@pytest.mark.asyncio
async def test_lock_context_manager_enter(mocker: MockerFixture) -> None:
    lock = aiofreqlimit.Lock()

    acquire = mocker.patch.object(lock._lock, "acquire")
    acquire.side_effect = RuntimeError("acquire")
    release = mocker.patch.object(lock._lock, "release")
    flag = False
    with pytest.raises(RuntimeError, match="acquire"):
        async with lock:
            flag = True
    assert lock.count == 1
    acquire.assert_called_once_with()
    release.assert_not_called()
    assert not flag


@pytest.mark.asyncio
async def test_lock_context_manager_exit(mocker: MockerFixture) -> None:
    lock = aiofreqlimit.Lock()

    acquire = mocker.patch.object(lock._lock, "acquire")
    release = mocker.patch.object(lock._lock, "release")
    release.side_effect = RuntimeError("release")
    flag = False
    with pytest.raises(RuntimeError, match="release"):
        async with lock:
            flag = True
    assert lock.count == 0
    acquire.assert_called_once_with()
    release.assert_called_once_with()
    assert flag


@pytest.mark.asyncio
async def test_lock_context_manager() -> None:
    lock = aiofreqlimit.Lock()

    async with lock:
        assert lock.count == 1
    assert lock.count == 0

    with suppress(RuntimeError):
        async with lock:
            assert lock.count == 1
            raise RuntimeError()
    assert lock.count == 0


@given(interval=floats(max_value=0))
def test_freq_limit_interval(interval: float) -> None:
    with pytest.raises(RuntimeError, match="Interval must be greater than 0"):
        aiofreqlimit.FreqLimit(interval)


@given(
    interval=floats(min_value=0, exclude_min=True),
    clean_interval=floats(max_value=0, exclude_max=True),
)
def test_freq_limit_clean_interval(
    interval: float, clean_interval: float
) -> None:
    with pytest.raises(
        RuntimeError,
        match="Clean interval must be greater than or equal to 0",
    ):
        aiofreqlimit.FreqLimit(interval, clean_interval)


@pytest.mark.asyncio
async def test_freq_limit() -> None:
    freq_limit = aiofreqlimit.FreqLimit(0.1)
    loop = asyncio.get_running_loop()
    time_marks = Tuple[float, float, float]

    async def limit(
        _freq_limit: aiofreqlimit.FreqLimit, interval: float
    ) -> time_marks:
        time1 = loop.time()
        async with _freq_limit.resource("key"):
            assert tuple(freq_limit._locks) == ("key",)
            time2 = loop.time()
            await asyncio.sleep(interval)
            time3 = loop.time()
        return time2, time3, time2 - time1

    tasks = (limit(freq_limit, uniform(0, 0.1)) for _ in range(5))
    intervals = cast(List[time_marks], await asyncio.gather(*tasks))
    assert all(isinstance(value, tuple) for value in intervals)
    intervals = sorted(intervals, key=lambda interval: interval[0])
    for i in range(len(intervals)):
        if i + 1 < len(intervals):
            assert intervals[i + 1][0] - intervals[i][0] > 0.1
            assert intervals[i][1] < intervals[i + 1][0]

    await asyncio.sleep(0.11)
    assert tuple(freq_limit._locks) == ()

    async with freq_limit.resource("key"):
        pass
    assert tuple(freq_limit._locks) == ("key",)
    await asyncio.sleep(0.33)
    assert tuple(freq_limit._locks) == ()

    await freq_limit.clear()
    assert freq_limit._clean_task is None
    assert tuple(freq_limit._locks) == ()
    assert not freq_limit._clean_event.is_set()


@pytest.mark.asyncio
async def test_freq_limit_keys() -> None:
    freq_limit = aiofreqlimit.FreqLimit(0.1)
    assert tuple(freq_limit._locks) == ()
    async with freq_limit.resource("key2"):
        assert tuple(freq_limit._locks.keys()) == ("key2",)
        async with freq_limit.resource("key3"):
            assert tuple(freq_limit._locks.keys()) == ("key2", "key3")
            async with freq_limit.resource("key4"):
                assert tuple(freq_limit._locks.keys()) == (
                    "key2",
                    "key3",
                    "key4",
                )
        await asyncio.sleep(0.11)
        assert tuple(freq_limit._locks.keys()) == ("key2",)
    await asyncio.sleep(0.11)
    assert tuple(freq_limit._locks) == ()
    await freq_limit.clear()


@pytest.mark.asyncio
async def test_freq_limit_overlaps() -> None:
    async def task1(_freq_limit: aiofreqlimit.FreqLimit) -> None:
        async with _freq_limit.resource("key1"):
            assert tuple(_freq_limit._locks) == ("key1",)
            await asyncio.sleep(0.11)
            assert tuple(_freq_limit._locks) == ("key1", "key2")

    async def task2(_freq_limit: aiofreqlimit.FreqLimit) -> None:
        await asyncio.sleep(0.05)
        assert tuple(_freq_limit._locks) == ("key1",)
        async with _freq_limit.resource("key2"):
            assert tuple(_freq_limit._locks) == ("key1", "key2")
            await asyncio.sleep(0.16)
            assert tuple(_freq_limit._locks) == ("key2", "key3")

    async def task3(_freq_limit: aiofreqlimit.FreqLimit) -> None:
        await asyncio.sleep(0.21)
        assert tuple(_freq_limit._locks) == ("key2",)
        async with _freq_limit.resource("key3"):
            assert tuple(_freq_limit._locks) == ("key2", "key3")
            lock2_ref = ref(_freq_limit._locks["key2"])
            async with _freq_limit.resource("key2"):
                assert tuple(_freq_limit._locks) == ("key2", "key3")
                assert lock2_ref() is _freq_limit._locks["key2"]
                await asyncio.sleep(0.1)
            assert tuple(_freq_limit._locks) == ("key2", "key3")
            await asyncio.sleep(0.1)
            assert lock2_ref() is None
            assert tuple(_freq_limit._locks) == ("key3",)
        assert tuple(_freq_limit._locks) == ("key3",)
        await asyncio.sleep(0.1)
        assert tuple(_freq_limit._locks) == ()

    freq_limit = aiofreqlimit.FreqLimit(0.1)
    await asyncio.gather(
        task1(freq_limit), task2(freq_limit), task3(freq_limit)
    )
    await freq_limit.clear()


@pytest.mark.asyncio
async def test_freq_limit_frequency() -> None:
    loop = asyncio.get_running_loop()
    intervals: List[float] = []
    time = loop.time()
    freq_limit = aiofreqlimit.FreqLimit(0.05)
    lock_ref = None
    for index in range(10):
        async with freq_limit.resource("key"):
            pass
        if index == 0:
            lock_ref = ref(freq_limit._locks["key"])
        else:
            intervals.append(loop.time() - time)
            assert lock_ref is not None
            assert lock_ref() is freq_limit._locks["key"]
        time = loop.time()
        assert "key" in freq_limit._locks
    assert all(0.05 <= interval <= 0.051 for interval in intervals)
    await freq_limit.clear()
